@@ -32,6 +32,8 @@
 #include <stdatomic.h>
 #include "librist/logging.h"
 
+#undef RIST_DEPRECATED
+
 #define UINT16_SIZE (UINT16_MAX + 1)
 // These 4 control the memory footprint and buffer capacity of the lib
 // They MUST be a power of two or wrap-around index calculations will break
@@ -45,6 +47,8 @@
 #define RIST_RTT_MIN (3)
 // this value is UINT32_MAX 4294967.296
 #define RIST_CLOCK (4294967LL)
+#define ONE_SECOND (1000 * RIST_CLOCK)
+#define RIST_LOG_QUIESCE_TIMER  ONE_SECOND
 /* nack requests are sent every time a data packet is received. */
 /* this timer will be triggered to ensure we output nacks even when there is no data coming in */
 #define RIST_MAX_JITTER (5) /* In milliseconds */
@@ -58,7 +62,6 @@
 // Maximum offset before the payload that the code can use to put in headers
 //#define RIST_MAX_PAYLOAD_OFFSET (sizeof(struct rist_gre_key_seq) + sizeof(struct rist_protocol_hdr))
 #define RIST_MAX_HEADER_SIZE 32
-
 #define CHECK_BIT(var,pos) !!((var) & (1<<(pos)))
 
 #define BYTE_TO_BINARY_PATTERN "%c%c%c%c%c%c%c%c"
@@ -104,6 +107,7 @@ struct rist_buffer {
 	struct rist_buffer *next_free;
 	size_t alloc_size;
 	bool free;
+	bool retry_queued;
 };
 
 struct rist_missing_buffer {
@@ -179,7 +183,7 @@ struct nacks {
 };
 
 struct rist_flow {
-	volatile int shutdown;
+	atomic_int shutdown;
 	int max_output_jitter;
 
 	struct rist_buffer *receiver_queue[RIST_SERVER_QUEUE_BUFFERS]; /* output queue */
@@ -209,6 +213,7 @@ struct rist_flow {
 	uint32_t missing_counter_max;
 
 	uint32_t flow_id;
+	uint32_t flow_id_actual;
 	int dead;
 	struct rist_flow *next;
 	struct rist_peer **peer_lst;
@@ -254,6 +259,7 @@ struct rist_flow {
 	size_t dataout_fifo_queue_bytesize;
 	atomic_ulong dataout_fifo_queue_read_index;
 	atomic_ulong dataout_fifo_queue_write_index;
+	atomic_bool fifo_overflow;
 
 	/* Temporary buffer for grouping and sending nacks */
 	struct nacks nacks;
@@ -261,14 +267,15 @@ struct rist_flow {
 };
 
 struct rist_retry {
-	uint32_t seq;
 	struct rist_peer *peer;
 	uint64_t insert_time;
+	uint32_t seq;
+	bool active;//signal whether this retry has been consumed (false) or not
 };
 
 struct rist_common_ctx {
-	volatile int shutdown;
-	volatile bool startup_complete;
+	atomic_int shutdown;
+	atomic_bool startup_complete;
 
 	/* Used by logging */
 	intptr_t sender_id;
@@ -278,6 +285,7 @@ struct rist_common_ctx {
 
 	/* Flows */
 	struct rist_flow *FLOWS;
+	pthread_mutex_t flows_lock;
 
 	/* evsocket */
 	struct evsocket_ctx *evctx;
@@ -329,6 +337,7 @@ struct rist_common_ctx {
 
 	int (*stats_callback)(void *arg, const struct rist_stats *stats_container);
 	void *stats_callback_argument;
+	pthread_mutex_t stats_lock;
 
 	pthread_rwlock_t oob_queue_lock;
 	struct rist_buffer *oob_queue[RIST_OOB_QUEUE_BUFFERS]; /* oob queue */
@@ -342,6 +351,9 @@ struct rist_common_ctx {
 	/* Connection status callback */
 	connection_status_callback_t connection_status_callback;
 	void *connection_status_callback_argument;
+
+	rist_thread_callback_func_t thread_callback;
+	void *thread_callback_arg;
 };
 
 struct rist_receiver {
@@ -350,11 +362,12 @@ struct rist_receiver {
 	pthread_mutex_t mutex;
 
 	/* Receiver data callback */
-	receiver_data_callback_t receiver_data_callback;
+	receiver_data_callback2_t receiver_data_callback;
 	void *receiver_data_callback_argument;
 	int receiver_data_ready_notify_fd;
 
 	/* Receiver thread variables */
+	bool protocol_running;
 	pthread_t receiver_thread;
 
 	/* Reporting id */
@@ -380,6 +393,7 @@ struct rist_sender {
 	bool null_packet_suppression;
 
 	/* Sender thread variables */
+	bool protocol_running;
 	pthread_t sender_thread;
 	/* data/nacks out thread signaling */
 	pthread_cond_t condition;
@@ -558,18 +572,21 @@ struct rist_peer {
 	struct rist_bandwidth_estimation retry_bw;
 
 	/* shutting down flag */
-	volatile bool shutdown;
+	atomic_bool shutdown;
 
 	/* Timers */
 	uint32_t rtcp_keepalive_interval;
 	uint64_t keepalive_next_time;
-	uint32_t session_timeout;
+	uint64_t session_timeout;
 	uint64_t last_rtcp_received;
 	uint64_t last_sender_report_time;
 	uint64_t last_sender_report_ts;
 
 	char *url;
 	char cname[RIST_MAX_HOSTNAME];
+	bool send_first_connection_event;
+
+	uint64_t log_repeat_timer;
 };
 
 static inline struct rist_common_ctx *rist_struct_get_common(struct rist_ctx *ctx) {
@@ -643,8 +660,11 @@ static inline void peer_append(struct rist_peer *p)
 	if (p->parent)
 	{
 		struct rist_peer *peer = p->parent;
-		if (!peer->child)
+		if (!peer->child) {
 			peer->child = p;
+			if (peer->sender_ctx)
+				peer->sender_ctx->total_weight += p->config.weight;
+		}
 		else
 		{
 			struct rist_peer *child = peer->child;
