@@ -11,6 +11,7 @@
 #include "log-private.h"
 #include "udp-private.h"
 #include "vcs_version.h"
+#include "rist-thread.h"
 #include <librist/version.h>
 #include "crypto/crypto-private.h"
 #include <assert.h>
@@ -18,6 +19,8 @@
 #include <processthreadsapi.h>
 #endif
 
+#undef RIST_DEPRECATED
+
 /* Receiver functions */
 
 int rist_receiver_create(struct rist_ctx **_ctx, enum rist_profile profile,
@@ -114,6 +117,7 @@ static struct rist_flow *rist_get_longest_flow(struct rist_receiver *ctx, ssize_
 	// Select the flow with highest queue count
 	ssize_t num_loop = 0;
 	struct rist_flow *f = NULL;
+	pthread_mutex_lock(&ctx->common.flows_lock);
 	struct rist_flow *f_loop = ctx->common.FLOWS;
 	while (f_loop) {
 		struct rist_flow *nextflow = f_loop->next;
@@ -128,10 +132,16 @@ static struct rist_flow *rist_get_longest_flow(struct rist_receiver *ctx, ssize_
 		}
 		f_loop = nextflow;
 	}
+	pthread_mutex_unlock(&ctx->common.flows_lock);
 	return f;
 }
 
-int rist_receiver_data_read(struct rist_ctx *rist_ctx, const struct rist_data_block **data_buffer, int timeout)
+int rist_receiver_data_read(struct rist_ctx *ctx, const struct rist_data_block **data_block, int timeout)
+{
+	return rist_receiver_data_read2(ctx, (struct rist_data_block **)data_block, timeout);
+}
+
+int rist_receiver_data_read2(struct rist_ctx *rist_ctx, struct rist_data_block **data_buffer, int timeout)
 {
 	if (RIST_UNLIKELY(!rist_ctx))
 	{
@@ -146,12 +156,12 @@ int rist_receiver_data_read(struct rist_ctx *rist_ctx, const struct rist_data_bl
 
 	struct rist_receiver *ctx = rist_ctx->receiver_ctx;
 
-	const struct rist_data_block *data_block = NULL;
+	struct rist_data_block *data_block = NULL;
 	/* We could enter the lock now, to read the counter. However performance penalties apply.
 	   The risks for not entering the lock are either sleeping too much (a packet gets added while we read)
 	   or not at all when we should (i.e.: the calling application is reading from multiple threads). Both
 	   risks are tolerable */
-	
+
 	ssize_t num = 0;
 	// Select the flow with highest queue count to minimize jitter for calling app
 	struct rist_flow *f = rist_get_longest_flow(ctx, &num);
@@ -188,10 +198,22 @@ int rist_receiver_data_read(struct rist_ctx *rist_ctx, const struct rist_data_bl
 
 	*data_buffer = data_block;
 
+	bool overflow = false;
+	while (!atomic_compare_exchange_weak(&f->fifo_overflow, &overflow, false)) {
+		//
+	}
+	if (overflow)
+		data_block->flags |= RIST_DATA_FLAGS_OVERFLOW;
+
 	return (int)num;
 }
 
-void rist_receiver_data_block_free(struct rist_data_block **block)
+void rist_receiver_data_block_free(struct rist_data_block **const block)
+{
+	rist_receiver_data_block_free2((struct rist_data_block **)block);
+}
+
+void rist_receiver_data_block_free2(struct rist_data_block **block)
 {
 	struct rist_data_block *b = *block;
 	if (b->ref != NULL)
@@ -247,24 +269,30 @@ int rist_connection_status_callback_set(struct rist_ctx *ctx, connection_status_
 	return 0;
 }
 
-int rist_receiver_data_callback_set(struct rist_ctx *rist_ctx,
-									int (*data_callback)(void *arg, const struct rist_data_block *data_block),
-									void *arg)
+int rist_receiver_data_callback_set(struct rist_ctx *rist_ctx,receiver_data_callback_t data_callback, void *arg)
 {
-	if (RIST_UNLIKELY(!rist_ctx))
-	{
-		rist_log_priv3(RIST_LOG_ERROR, "ctx is null on rist_receiver_data_callback_set call!\n");
-		return -1;
-	}
-	if (RIST_UNLIKELY(rist_ctx->mode != RIST_RECEIVER_MODE || !rist_ctx->receiver_ctx))
-	{
-		rist_log_priv3(RIST_LOG_ERROR, "rist_receiver_data_callback_set call with CTX not set up for receiving\n");
-		return -1;
-	}
-	struct rist_receiver *ctx = rist_ctx->receiver_ctx;
-	ctx->receiver_data_callback = data_callback;
-	ctx->receiver_data_callback_argument = arg;
-	return 0;
+  return rist_receiver_data_callback_set2(
+      rist_ctx, (receiver_data_callback2_t)data_callback, arg);
+}
+
+int rist_receiver_data_callback_set2(struct rist_ctx *rist_ctx,
+                                     receiver_data_callback2_t data_callback,
+                                     void *arg) {
+  if (RIST_UNLIKELY(!rist_ctx)) {
+    rist_log_priv3(RIST_LOG_ERROR,
+                   "ctx is null on rist_receiver_data_callback_set call!\n");
+    return -1;
+  }
+  if (RIST_UNLIKELY(rist_ctx->mode != RIST_RECEIVER_MODE ||
+                    !rist_ctx->receiver_ctx)) {
+    rist_log_priv3(RIST_LOG_ERROR, "rist_receiver_data_callback_set call with "
+                                   "CTX not set up for receiving\n");
+    return -1;
+  }
+  struct rist_receiver *ctx = rist_ctx->receiver_ctx;
+  ctx->receiver_data_callback = data_callback;
+  ctx->receiver_data_callback_argument = arg;
+  return 0;
 }
 
 /* Sender functions */
@@ -381,13 +409,6 @@ int rist_sender_create(struct rist_ctx **_ctx, enum rist_profile profile,
 
 	ctx->sender_initialized = true;
 
-	if (pthread_create(&ctx->sender_thread, NULL, sender_pthread_protocol, (void *)ctx) != 0)
-	{
-		rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Could not created sender thread.\n");
-		ret = -3;
-		goto free_ctx_and_ret;
-	}
-
 	*_ctx = rist_ctx;
 	return 0;
 
@@ -622,6 +643,11 @@ int rist_stats_free(const struct rist_stats *stats_container)
 }
 
 int rist_peer_config_free(const struct rist_peer_config **peer_config)
+{
+  return rist_peer_config_free2((struct rist_peer_config **)peer_config);
+}
+
+int rist_peer_config_free2(struct rist_peer_config **peer_config)
 {
 	if (*peer_config) {
 		free((void *)*peer_config);
@@ -631,6 +657,11 @@ int rist_peer_config_free(const struct rist_peer_config **peer_config)
 }
 
 int rist_logging_settings_free(const struct rist_logging_settings **logging_settings)
+{
+	return rist_logging_settings_free2((struct rist_logging_settings **)logging_settings);
+}
+
+int rist_logging_settings_free2(struct rist_logging_settings **logging_settings)
 {
 	if (*logging_settings) {
 		free((void *)*logging_settings);
@@ -640,6 +671,11 @@ int rist_logging_settings_free(const struct rist_logging_settings **logging_sett
 }
 
 int rist_udp_config_free(const struct rist_udp_config **udp_config)
+{
+	return rist_udp_config_free2((struct rist_udp_config **)udp_config);
+}
+
+int rist_udp_config_free2(struct rist_udp_config **udp_config)
 {
 	if (*udp_config) {
 		free((void *)*udp_config);
@@ -656,6 +692,7 @@ int rist_stats_callback_set(struct rist_ctx *ctx, int statsinterval, int (*stats
 		return -1;
 	}
 	struct rist_common_ctx *cctx = rist_struct_get_common(ctx);
+	pthread_mutex_lock(&cctx->stats_lock);
 	if (RIST_UNLIKELY(!cctx))
 		return -1;
 
@@ -674,7 +711,7 @@ int rist_stats_callback_set(struct rist_ctx *ctx, int statsinterval, int (*stats
 			}
 		}
 	}
-
+	pthread_mutex_unlock(&cctx->stats_lock);
 	return 0;
 }
 
@@ -690,6 +727,11 @@ const char *librist_api_version(void)
 }
 
 int rist_parse_udp_address(const char *url, const struct rist_udp_config **udp_config)
+{
+  	return rist_parse_udp_address2(url, (struct rist_udp_config **)udp_config);
+}
+
+int rist_parse_udp_address2(const char *url, struct rist_udp_config **udp_config)
 {
 
 	int ret = 0;
@@ -737,6 +779,11 @@ int rist_peer_config_defaults_set(struct rist_peer_config *peer_config)
 }
 
 int rist_parse_address(const char *url, const struct rist_peer_config **peer_config)
+{
+	return rist_parse_address2(url, (struct rist_peer_config **)peer_config);
+}
+
+int rist_parse_address2(const char *url, struct rist_peer_config **peer_config)
 {
 
 	char * url_local = strdup(url);
@@ -895,6 +942,49 @@ int rist_peer_create(struct rist_ctx *ctx, struct rist_peer **peer, const struct
 	return ret;
 }
 
+int rist_peer_weight_set(struct rist_ctx *ctx, struct rist_peer *peer, const uint32_t weight)
+{
+	if (!ctx)
+	{
+		rist_log_priv3(RIST_LOG_ERROR, "rist_peer_weight_set call with null ctx\n");
+		return -1;
+	}
+
+	if (peer->parent != NULL)
+	{
+		rist_log_priv3(RIST_LOG_ERROR, "rist_peer_weight_set cannot be applied to peer with parent\n");
+		return -1;
+	}
+
+	if (ctx->mode == RIST_RECEIVER_MODE && ctx->receiver_ctx)
+	{
+		struct rist_receiver *rctx = ctx->receiver_ctx;
+		pthread_mutex_lock(&rctx->common.peerlist_lock);
+		peer->config.weight = weight;
+		peer->w_count = weight;
+		pthread_mutex_unlock(&rctx->common.peerlist_lock);
+	}
+	else if (ctx->mode == RIST_SENDER_MODE && ctx->sender_ctx)
+	{
+		struct rist_sender *sctx = ctx->sender_ctx;
+		pthread_mutex_lock(&sctx->mutex);
+		pthread_mutex_lock(&sctx->common.peerlist_lock);
+		uint32_t cur_weight = peer->config.weight;
+		peer->config.weight = weight;
+		if ((peer->listening && peer->child != NULL) || !peer->listening) {
+			sctx->total_weight -= cur_weight;
+			peer->w_count = peer->config.weight;
+			sctx->total_weight += peer->config.weight;
+		}
+		pthread_mutex_unlock(&sctx->common.peerlist_lock);
+		pthread_mutex_unlock(&sctx->mutex);
+	}
+	else
+		return -1;
+
+	return 0;
+}
+
 int rist_peer_destroy(struct rist_ctx *ctx, struct rist_peer *peer) {
 	if (!ctx) {
 		rist_log_priv3(RIST_LOG_ERROR, "rist_peer_destroy call with null ctx\n");
@@ -916,33 +1006,53 @@ int rist_peer_destroy(struct rist_ctx *ctx, struct rist_peer *peer) {
 
 static int rist_sender_start(struct rist_sender *ctx)
 {
-	if (!ctx->sender_initialized)
-	{
-		return -1;
+	pthread_mutex_lock(&ctx->mutex);
+	if (!ctx->protocol_running) {
+		if (rist_thread_create(&ctx->common, &ctx->sender_thread, NULL, sender_pthread_protocol, (void *)ctx) != 0)
+		{
+			rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Could not created sender thread.\n");
+			goto unlock_failed;
+		}
+		ctx->protocol_running = true;
+	} else {
+		goto unlock_failed;
 	}
-
 	if (ctx->total_weight > 0)
 	{
 		ctx->weight_counter = ctx->total_weight;
 		rist_log_priv(&ctx->common, RIST_LOG_INFO, "Total weight: %lu\n", ctx->total_weight);
 	}
+	atomic_store_explicit(&ctx->common.startup_complete, true, memory_order_release);
 
-	ctx->common.startup_complete = true;
+	pthread_mutex_unlock(&ctx->mutex);
 	return 0;
+
+unlock_failed:
+	pthread_mutex_unlock(&ctx->mutex);
+	return -1;
 }
 
 static int rist_receiver_start(struct rist_receiver *ctx)
 {
-	if (!ctx->receiver_thread)
+	pthread_mutex_lock(&ctx->mutex);
+	if (!ctx->protocol_running)
 	{
-		if (pthread_create(&ctx->receiver_thread, NULL, receiver_pthread_protocol, (void *)ctx) != 0)
+		if (rist_thread_create(&ctx->common, &ctx->receiver_thread, NULL, receiver_pthread_protocol, (void *)ctx) != 0)
 		{
 			rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Could not create receiver protocol thread.\n");
-			return -1;
+			goto unlock_failed;
 		}
+		ctx->protocol_running = true;
+	} else {
+		goto unlock_failed;
 	}
 
+	pthread_mutex_unlock(&ctx->mutex);
 	return 0;
+
+unlock_failed:
+	pthread_mutex_unlock(&ctx->mutex);
+	return -1;
 }
 
 int rist_start(struct rist_ctx *ctx) {
@@ -966,19 +1076,12 @@ static int rist_sender_destroy(struct rist_sender *ctx)
 	}
 
 	rist_log_priv(&ctx->common, RIST_LOG_INFO, "Triggering protocol loop termination\n");
-	ctx->common.shutdown = 1;
-	uint64_t start_time = timestampNTP_u64();
-	while (ctx->sender_thread && ctx->common.shutdown != 2)
-	{
-		rist_log_priv(&ctx->common, RIST_LOG_INFO, "Waiting for protocol loop to exit\n");
-		usleep(5000);
-		if (((timestampNTP_u64() - start_time) / RIST_CLOCK) > 10000)
-		{
-			rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Protocol loop took more than 10 seconds to exit. Something is wrong!\n");
-			assert(0);
-		}
-	}
-	pthread_join(ctx->sender_thread, NULL);
+	atomic_store_explicit(&ctx->common.shutdown, 1, memory_order_release);
+	pthread_mutex_lock(&ctx->mutex);
+	bool running = ctx->protocol_running;
+	pthread_mutex_unlock(&ctx->mutex);
+	if (running)
+		pthread_join(ctx->sender_thread, NULL);
 	rist_sender_destroy_local(ctx);
 
 	return 0;
@@ -992,19 +1095,12 @@ static int rist_receiver_destroy(struct rist_receiver *ctx)
 	}
 
 	rist_log_priv(&ctx->common, RIST_LOG_INFO, "Triggering protocol loop termination\n");
-	ctx->common.shutdown = 1;
-	uint64_t start_time = timestampNTP_u64();
-	while (ctx->receiver_thread && ctx->common.shutdown != 2)
-	{
-		rist_log_priv(&ctx->common, RIST_LOG_INFO, "Waiting for protocol loop to exit\n");
-		usleep(5000);
-		if (((timestampNTP_u64() - start_time) / RIST_CLOCK) > 10000)
-		{
-			rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Protocol loop took more than 10 seconds to exit. Something is wrong!\n");
-			assert(0);
-		}
-	}
-	pthread_join(ctx->receiver_thread, NULL);
+	atomic_store_explicit(&ctx->common.shutdown, 1, memory_order_release);
+	pthread_mutex_lock(&ctx->mutex);
+	bool running = ctx->protocol_running;
+	pthread_mutex_unlock(&ctx->mutex);
+	if (running)
+		pthread_join(ctx->receiver_thread, NULL);
 	rist_receiver_destroy_local(ctx);
 
 	return 0;
@@ -1032,7 +1128,7 @@ int rist_receiver_set_output_fifo_size(struct rist_ctx *ctx, uint32_t desired_si
 		rist_log_priv3(RIST_LOG_ERROR, "rist_receiver_set_fifo_size called with null ctx\n");
 		return -1;
 	}
-	if (ctx->mode == RIST_RECEIVER_MODE || !ctx->receiver_ctx)
+	if (ctx->mode != RIST_RECEIVER_MODE || !ctx->receiver_ctx)
 	{
 		rist_log_priv3(RIST_LOG_ERROR, "rist_receiver_set_fifo_size can only be called on receiver\n");
 		return -2;
@@ -1050,3 +1146,30 @@ int rist_receiver_set_output_fifo_size(struct rist_ctx *ctx, uint32_t desired_si
 	ctx->receiver_ctx->fifo_queue_size = desired_size;
 	return 0;
 }
+
+int rist_set_opt(struct rist_ctx *ctx, enum rist_opt opt, void* optval1, void* optval2, void* optval3)
+{
+	struct rist_common_ctx *cctx = NULL;
+	if (ctx->mode == RIST_RECEIVER_MODE && ctx->receiver_ctx != NULL)
+		cctx = &ctx->receiver_ctx->common;
+	else if (ctx->mode == RIST_SENDER_MODE && ctx->sender_ctx != NULL)
+		cctx = &ctx->sender_ctx->common;
+	else
+		return -1;
+	
+	switch(opt) {
+	case RIST_OPT_THREAD_CALLBACK:
+		;
+		rist_thread_callback_t *thread_callback  = optval1;
+		if (thread_callback == NULL || thread_callback->thread_callback == NULL || optval3 != NULL)
+			return -1;
+		if (atomic_load_explicit(&cctx->startup_complete, memory_order_acquire)) 
+			return -1;
+		cctx->thread_callback = thread_callback->thread_callback;
+		cctx->thread_callback_arg = optval2;
+		break;
+	default:
+		return -1;
+	}
+	return 0;
+}
