@@ -11,7 +11,9 @@
 #include "log-private.h"
 #include "socket-shim.h"
 #include "endian-shim.h"
+#if HAVE_MBEDTLS
 #include "eap.h"
+#endif
 #include "crypto/psk.h"
 #include "mpegts.h"
 #include <stdlib.h>
@@ -19,6 +21,7 @@
 #include <errno.h>
 #include <stdint.h>
 #include <assert.h>
+#include <fcntl.h>
 
 uint64_t timestampNTP_u64(void)
 {
@@ -44,7 +47,7 @@ uint64_t timestampNTP_u64(void)
 	t /= 1000000000;
 	// There is 70 years (incl. 17 leap ones) offset to the Unix Epoch.
 	// No leap seconds during that period since they were not invented yet.
-	t |= ((70LL * 365 + 17) * 24 * 60 * 60 + ts.tv_sec) << 32;
+	t |= (uint64_t)((70LL * 365 + 17) * 24 * 60 * 60 + ts.tv_sec) << 32;
 	return t; // nanoseconds (technically, 232.831 picosecond units)
 }
 
@@ -60,7 +63,7 @@ uint64_t timestampNTP_RTC_u64(void) {
 	t /= 1000000000;
 	// There is 70 years (incl. 17 leap ones) offset to the Unix Epoch.
 	// No leap seconds during that period since they were not invented yet.
-	t |= ((70LL * 365 + 17) * 24 * 60 * 60 + ts.tv_sec) << 32;
+	t |= (uint64_t)((70LL * 365 + 17) * 24 * 60 * 60 + ts.tv_sec) << 32;
 	return t;
 }
 
@@ -159,7 +162,7 @@ void rist_clean_sender_enqueue(struct rist_sender *ctx)
 
 }
 
-size_t rist_send_seq_rtcp(struct rist_peer *p, uint16_t seq_rtp, uint8_t payload_type, uint8_t *payload, size_t payload_len, uint64_t source_time, uint16_t src_port, uint16_t dst_port)
+size_t rist_send_seq_rtcp(struct rist_peer *p, uint16_t seq_rtp, uint8_t payload_type, uint8_t *payload, size_t payload_len, uint64_t source_time, uint16_t src_port, uint16_t dst_port, bool retry)
 {
 	struct rist_common_ctx *ctx = get_cctx(p);
 	struct rist_key *k = &p->key_tx;
@@ -227,7 +230,7 @@ size_t rist_send_seq_rtcp(struct rist_peer *p, uint16_t seq_rtp, uint8_t payload
 				SET_BIT(hdr->rtp.flags, 4);
 			hdr->rtp.ssrc = htobe32(p->adv_flow_id);
 			hdr->rtp.seq = htobe16(seq_rtp);
-			if ((seq_rtp + 1) != ctx->seq_rtp)
+			if (retry)
 			{
 				// This is a retransmission
 				//rist_log_priv(&ctx->common, RIST_LOG_ERROR, "\tResending: %"PRIu32"/%"PRIu16"/%"PRIu32"\n", seq, seq_rtp, ctx->seq);
@@ -326,7 +329,7 @@ int rist_send_common_rtcp(struct rist_peer *p, uint8_t payload_type, uint8_t *pa
 	if (RIST_UNLIKELY(p->config.timing_mode == RIST_TIMING_MODE_ARRIVAL) && !p->receiver_mode)
 		source_time = timestampNTP_u64();
 
-	size_t ret = rist_send_seq_rtcp(p, (uint16_t)seq_rtp, payload_type, payload, payload_len, source_time, src_port, dst_port);
+	size_t ret = rist_send_seq_rtcp(p, (uint16_t)seq_rtp, payload_type, payload, payload_len, source_time, src_port, dst_port, false);
 
 	if ((!p->compression && ret < payload_len) || ret <= 0)
 	{
@@ -489,7 +492,7 @@ void rist_create_socket(struct rist_peer *peer)
 		}
 
 		peer->sd = udpsocket_open_bind(host, port, &peer->miface[0]);
-		if (peer->sd > 0) {
+		if (peer->sd >= 0) {
 			rist_log_priv(get_cctx(peer), RIST_LOG_INFO, "Starting in URL listening mode (socket# %d)\n", peer->sd);
 		} else {
 			rist_log_priv(get_cctx(peer), RIST_LOG_ERROR, "Could not start in URL listening mode. %s\n", strerror(errno));
@@ -515,7 +518,7 @@ void rist_create_socket(struct rist_peer *peer)
 		// We use sendto ... so, no need to connect directly here
 		peer->sd = udpsocket_open(peer->address_family);
 		// TODO : set max hops
-		if (peer->sd > 0)
+		if (peer->sd >= 0)
 			rist_log_priv(get_cctx(peer), RIST_LOG_INFO, "Starting in URL connect mode (%d)\n", peer->sd);
 		else {
 			rist_log_priv(get_cctx(peer), RIST_LOG_ERROR, "Could not start in URL connect mode. %s\n", strerror(errno));
@@ -545,7 +548,13 @@ void rist_create_socket(struct rist_peer *peer)
 	if (peer->cname[0] == 0)
 		rist_populate_cname(peer);
 	rist_log_priv(get_cctx(peer), RIST_LOG_INFO, "Peer cname is %s\n", peer->cname);
-
+#ifndef _WIN32
+	if (fcntl(peer->sd, F_SETFD, FD_CLOEXEC) == -1) {
+		udpsocket_close(peer->sd);
+		rist_log_priv(get_cctx(peer), RIST_LOG_ERROR, "Could not set close-on-exec\n");
+		peer->sd = -1;
+	}
+#endif
 }
 
 static inline void rist_rtcp_write_empty_rr(uint8_t *buf, int *offset, const uint32_t flow_id) {
@@ -874,6 +883,7 @@ void rist_sender_send_data_balanced(struct rist_sender *ctx, struct rist_buffer
 
 	//We can do it safely here, since this function is only to be called once per packet
 	buffer->seq = ctx->common.seq++;
+	uint64_t now = timestampNTP_u64();
 
 peer_select:
 
@@ -882,7 +892,7 @@ peer_select:
 
 		if (!peer->is_data || peer->parent)
 			continue;
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 		if (!peer->listening && !eap_is_authenticated(peer->eap_ctx))
 			continue;
 #endif
@@ -907,19 +917,19 @@ peer_select:
 			if (peer->listening) {
 				struct rist_peer *child = peer->child;
 				while (child) {
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 					if (!eap_is_authenticated(child->eap_ctx))
 					{
 						//do nothing
 					} else
 #endif
-					if (child->is_data && !child->dead) {
-					uint8_t *payload = buffer->data;
-					rist_send_common_rtcp(child, buffer->type, &payload[RIST_MAX_PAYLOAD_OFFSET], buffer->size, buffer->source_time, buffer->src_port, buffer->dst_port, buffer->seq_rtp);
+					if (child->is_data && (!child->dead || (child->dead && (child->dead_since + peer->recovery_buffer_ticks) < now))) {
+						uint8_t *payload = buffer->data;
+						rist_send_common_rtcp(child, buffer->type, &payload[RIST_MAX_PAYLOAD_OFFSET], buffer->size, buffer->source_time, buffer->src_port, buffer->dst_port, buffer->seq_rtp);
 					}
 					child = child->sibling_next;
 				}
-			} else {
+			} else if (!peer->dead || (peer->dead && (peer->dead_since + peer->recovery_buffer_ticks) < now)) {
 				uint8_t *payload = buffer->data;
 				rist_send_common_rtcp(peer, buffer->type, &payload[RIST_MAX_PAYLOAD_OFFSET], buffer->size, buffer->source_time, buffer->src_port, buffer->dst_port, buffer->seq_rtp);
 			}
@@ -934,41 +944,38 @@ peer_select:
 		}
 	}
 	looped = true;
-
 	if (selected_peer_by_weight) {
 		peer = selected_peer_by_weight;
 		if (peer->listening) {
 			struct rist_peer *child = peer->child;
 			while (child) {
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 					if (!eap_is_authenticated(child->eap_ctx))
 					{
 						//do nothing
 					} else
 #endif
-				if (child->is_data && !child->dead) {
+				if (child->is_data && (!child->dead || (child->dead && (child->dead_since + peer->recovery_buffer_ticks) < now))) {
 					uint8_t *payload = buffer->data;
 					rist_send_common_rtcp(child, buffer->type, &payload[RIST_MAX_PAYLOAD_OFFSET], buffer->size, buffer->source_time, buffer->src_port, buffer->dst_port,  buffer->seq_rtp);
 				}
 				child = child->sibling_next;
 			}
-		} else {
+		} else if (!peer->dead || (peer->dead && (peer->dead_since + peer->recovery_buffer_ticks) < now)) {
 			uint8_t *payload = buffer->data;
 			rist_send_common_rtcp(peer, buffer->type, &payload[RIST_MAX_PAYLOAD_OFFSET], buffer->size, buffer->source_time, buffer->src_port, buffer->dst_port, buffer->seq_rtp);
-			ctx->weight_counter--;
-			peer->w_count--;
 		}
+		ctx->weight_counter--;
+		peer->w_count--;
 	}
 
 	if (ctx->total_weight > 0 && (ctx->weight_counter == 0 || !selected_peer_by_weight)) {
 		peer = ctx->common.PEERS;
 		ctx->weight_counter = ctx->total_weight;
 		for (; peer; peer = peer->next) {
-			if (peer->listening || !peer->is_data)
-				continue;
 			peer->w_count = peer->config.weight;
 		}
-		if (!selected_peer_by_weight && peercnt > 0)
+		if (!looped && !selected_peer_by_weight && peercnt > 0)
 			goto peer_select;
 	}
 }
@@ -989,17 +996,9 @@ size_t rist_get_sender_retry_queue_size(struct rist_sender *ctx)
 /* This function must return, 0 when there is nothing to send, < 0 on error and > 0 for bytes sent */
 ssize_t rist_retry_dequeue(struct rist_sender *ctx)
 {
-//	rist_log_priv(&ctx->common, RIST_LOG_ERROR,
-//			"\tCurrent read/write index are %zu/%zu \n", ctx->sender_retry_queue_read_index,
-//			ctx->sender_retry_queue_write_index);
-
-	// TODO: Is this logic flawed and we are always one unit behind (look at oob_dequee)
 	size_t sender_retry_queue_read_index = (ctx->sender_retry_queue_read_index + 1)& (ctx->sender_retry_queue_size -1);
 
 	if (sender_retry_queue_read_index == ctx->sender_retry_queue_write_index) {
-		//rist_log_priv(&ctx->common, RIST_LOG_ERROR,
-		//	"\t[GOOD] We are all up to date, index is %" PRIu64 "\n",
-		//	ctx->sender_retry_queue_read_index);
 		return 0;
 	}
 
@@ -1008,16 +1007,16 @@ ssize_t rist_retry_dequeue(struct rist_sender *ctx)
 
 	// If they request a non-sense seq number, we will catch it when we check the seq number against
 	// the one on that buffer position and it does not match
-
+	
 	size_t idx = rist_sender_index_get(ctx, retry->seq);
-	if (ctx->sender_queue[idx] == NULL) {
+	if (RIST_UNLIKELY(ctx->sender_queue[idx] == NULL)) {
 		rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 			" Couldn't find block %" PRIu32 " (i=%zu/r=%zu/w=%zu/d=%zu/rs=%zu), consider increasing the buffer size\n",
 			retry->seq, idx, atomic_load_explicit(&ctx->sender_queue_read_index, memory_order_acquire), atomic_load_explicit(&ctx->sender_queue_write_index, memory_order_acquire), ctx->sender_queue_delete_index,
 			rist_get_sender_retry_queue_size(ctx));
 		retry->peer->stats_sender_instant.retrans_skip++;
 		return -1;
-	} else if ((uint16_t)retry->seq != ctx->sender_queue[idx]->seq_rtp) {
+	} else if (RIST_UNLIKELY((uint16_t)retry->seq != ctx->sender_queue[idx]->seq_rtp)) {
 		rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 			" Couldn't find block %" PRIu16 " (i=%zu/r=%zu/w=%zu/d=%zu/rs=%zu), found an old one instead %" PRIu32 " (%" PRIu64 "), bitrate is too high\n",
 			(uint16_t)retry->seq, idx, atomic_load_explicit(&ctx->sender_queue_read_index, memory_order_acquire), atomic_load_explicit(&ctx->sender_queue_write_index, memory_order_acquire), ctx->sender_queue_delete_index,
@@ -1025,6 +1024,9 @@ ssize_t rist_retry_dequeue(struct rist_sender *ctx)
 		retry->peer->stats_sender_instant.retrans_skip++;
 		return -1;
 	}
+	/* we're consuming the retry for an existing buffer, set it to false to allow new retries to come in */
+	ctx->sender_queue[idx]->retry_queued = false;
+	retry->active = false;
 
 	// TODO: re-enable rist_send_data_allowed (cooldown feature)
 
@@ -1063,9 +1065,10 @@ ssize_t rist_retry_dequeue(struct rist_sender *ctx)
 
 	// Check buffer element age
 	uint64_t now = timestampNTP_u64();
+	/* queue_time holds the original insertion time for this seq */
 	uint64_t data_age = (now - ctx->sender_queue[idx]->time) / RIST_CLOCK;
 	uint64_t retry_age = (now - retry->insert_time) / RIST_CLOCK;
-	if (retry_age > retry->peer->config.recovery_length_max) {
+	if (RIST_UNLIKELY(retry_age > retry->peer->config.recovery_length_max)) {
 		rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 			"Retry-request of element %" PRIu32 " (idx %zu) that was sent %" PRIu64
 				"ms ago has been in the queue too long to matter: %"PRIu64"ms > %ums\n",
@@ -1075,7 +1078,6 @@ ssize_t rist_retry_dequeue(struct rist_sender *ctx)
 	}
 
 	struct rist_buffer *buffer = ctx->sender_queue[idx];
-	/* queue_time holds the original insertion time for this seq */
 	if (ctx->common.debug)
 		rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 			"Resending %"PRIu32"/%"PRIu32"/%"PRIu16" (idx %zu) after %" PRIu64
@@ -1085,46 +1087,34 @@ ssize_t rist_retry_dequeue(struct rist_sender *ctx)
 
 	uint8_t *payload = buffer->data;
 
-	// TODO: I do not think this check is needed anymore ... we fixed the bug that was causing
-	// this scenario ... and we have thread-locking to prevent this
-	/*
-	if (!payload)
-	{
-		rist_log_priv(&ctx->common, RIST_LOG_ERROR,
-			"Someone deleted my buffer when resending %" PRIu32 " (idx %zu) after %" PRIu64
-			"ms of first transmission and %"PRIu64"ms in queue, bitrate is %zu + %zu, %zu\n",
-			retry->seq, idx, data_age, retry_age, retry->peer->bw.bitrate, retry_bw->bitrate,
-			retry->peer->bw.bitrate + retry_bw->bitrate);
-	}
-	*/
-
-	buffer->transmit_count++;
 	size_t ret = 0;
 	if (buffer->transmit_count >= retry->peer->config.max_retries) {
 		rist_log_priv(&ctx->common, RIST_LOG_ERROR, "Datagram %"PRIu32
 			" is missing, but nack count is too large (%u), age is %"PRIu64"ms, retry #%lu\n",
-			buffer->seq, buffer->transmit_count, data_age, buffer->transmit_count);
-	}
-	else {
-		uint16_t src_port = buffer->src_port;
-		if (src_port == 0)
-			src_port = 32768 + retry->peer->peer_data->adv_peer_id;
-		ret = (size_t)rist_send_seq_rtcp(retry->peer->peer_data, buffer->seq_rtp, buffer->type, &payload[RIST_MAX_PAYLOAD_OFFSET], buffer->size, buffer->source_time, src_port, retry->peer->peer_data->config.virt_dst_port);
-		// update bandwidth value
-		rist_calculate_bitrate(ret, retry_bw);
+			retry->seq, buffer->transmit_count, data_age, buffer->transmit_count);
+			retry->peer->stats_sender_instant.retrans_skip++;
+			return -1;
 	}
 
+	uint16_t src_port = buffer->src_port;
+	if (src_port == 0)
+		src_port = 32768 + retry->peer->peer_data->adv_peer_id;
+	ret = (size_t)rist_send_seq_rtcp(retry->peer->peer_data, buffer->seq_rtp, buffer->type, &payload[RIST_MAX_PAYLOAD_OFFSET], buffer->size, buffer->source_time, src_port, (retry->peer->peer_data->config.virt_dst_port & ~1UL), true);
+	// update bandwidth value
+	rist_calculate_bitrate(ret, retry_bw);
+
 	if (ret < buffer->size) {
 		rist_log_priv(&ctx->common, RIST_LOG_ERROR,
-			"Resending of packet failed %zu != %zu for seq %"PRIu32"\n", ret, buffer->size, buffer->seq);
+			"Resending of packet failed %zu != %zu for seq %"PRIu32"\n", ret, buffer->size, retry->seq);
 		retry->peer->stats_sender_instant.retrans_skip++;
-	} else {
-		if (retry->peer->peer_data)
-			retry->peer->peer_data->stats_sender_instant.retrans++;
-		else
-			retry->peer->stats_sender_instant.retrans++;
+		return -1;
 	}
 
+	buffer->transmit_count++;
+	if (retry->peer->peer_data)
+		retry->peer->peer_data->stats_sender_instant.retrans++;
+	else
+		retry->peer->stats_sender_instant.retrans++;
 	return ret;
 }
 
@@ -1138,8 +1128,9 @@ void rist_retry_enqueue(struct rist_sender *ctx, uint32_t seq, struct rist_peer
 	// Even though all the checks are on the dequeue function, we leave one here
 	// to prevent the flooding of our fifo .. It is based on the date of the
 	// last queued item with the same seq for this peer.
-	// The policy of whether to allow or not allow duplicate seq entries in the retry queue
+	// The policy of whether to allow or not allow duplicate inactive seq entries in the retry queue
 	// is dependent on the bloat_mode.
+	// No duplicate unhandled (i.e.: still queued) retries are accepted.
 	// bloat_mode disabled mode = unlimited duplicates
 	// bloat_mode normal mode = we enforce rtt spacing and allow duplicates
 	// bloat_mode aggresive mode = we enforce 2*rtt spacing and allow duplicates
@@ -1165,8 +1156,11 @@ void rist_retry_enqueue(struct rist_sender *ctx, uint32_t seq, struct rist_peer
 			if (ctx->common.debug)
 				rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 					"Nack request for seq %" PRIu32 " with age %" PRIu64 "ms and rtt_min %" PRIu32 " for peer #%d\n",
-					buffer->seq, age_ticks / RIST_CLOCK, peer->config.recovery_rtt_min, peer->adv_peer_id);
+					seq, age_ticks / RIST_CLOCK, peer->config.recovery_rtt_min, peer->adv_peer_id);
 		} else if (ctx->peer_lst_len == 1) {
+			/* there is a retry outstanding for this buffer, no need to add another */
+			if (buffer->retry_queued)
+				return;
 			// Only one peer (faster algorithm with no lookups)
 			if (buffer->last_retry_request != 0)
 			{
@@ -1176,7 +1170,7 @@ void rist_retry_enqueue(struct rist_sender *ctx, uint32_t seq, struct rist_peer
 				if (ctx->common.debug)
 					rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 						"Nack request for seq %" PRIu32 " with delta %" PRIu64 "ms, age %" PRIu64 "ms and rtt_min %" PRIu32 "\n",
-						buffer->seq, delta, age_ticks / RIST_CLOCK, peer->config.recovery_rtt_min);
+						seq, delta, age_ticks / RIST_CLOCK, peer->config.recovery_rtt_min);
 				uint64_t rtt = peer->last_mrtt;
 				if (peer->config.recovery_rtt_min > rtt)
 					rtt = peer->config.recovery_rtt_min;
@@ -1190,22 +1184,23 @@ void rist_retry_enqueue(struct rist_sender *ctx, uint32_t seq, struct rist_peer
 				{
 					rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 						"Nack request for seq %" PRIu32 ", age %"PRIu64"ms, is already queued (too soon to add another one), skipped, %" PRIu64 " < %" PRIu64 " ms\n",
-						buffer->seq, age_ticks / RIST_CLOCK, delta, rtt);
+						seq, age_ticks / RIST_CLOCK, delta, rtt);
 					peer->stats_sender_instant.bloat_skip++;
 					return;
 				}
+				buffer->retry_queued = true;
 			}
 			else
 			{
 				if (ctx->common.debug)
 					rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
-						"First nack request for seq %"PRIu32", age %"PRIu64"ms\n", buffer->seq, age_ticks / RIST_CLOCK);
+						"First nack request for seq %"PRIu32", age %"PRIu64"ms\n", seq, age_ticks / RIST_CLOCK);
 			}
 		} else {
 			// Multiple peers, we need to search for other retries in the queue for comparison
 			uint64_t delta = 0;
 			//We work backwards from the write index till we either find a retry with same peer & seq
-			//or it's too old to matter (older than 1 RTT ago)
+			//or it's too old to matter,looking up to 8 RTT's ago (4 in normal mode, 8 in aggressive)
 			size_t index = (ctx->sender_retry_queue_write_index -1) & (ctx->sender_retry_queue_size -1);
 			uint64_t rtt = peer->last_mrtt;
 			if (peer->config.recovery_length_min > rtt)
@@ -1214,22 +1209,26 @@ void rist_retry_enqueue(struct rist_sender *ctx, uint32_t seq, struct rist_peer
 			if (peer->config.congestion_control_mode == RIST_CONGESTION_CONTROL_MODE_AGGRESSIVE)
 				rtt *= 2;
 			struct rist_retry *lookup = NULL;
+			uint64_t search_period = rtt * 4;
 			while (index != ctx->sender_retry_queue_write_index) {
 				lookup = &ctx->sender_retry_queue[index];
 				if (lookup->seq == seq && lookup->peer == peer)
 					break;
-				if (lookup->insert_time < (now - rtt))
+				if (lookup->insert_time < (now - search_period))
 					break;
 				index = (index -1 ) & (ctx->sender_retry_queue_size -1);
 			}
 			retry = &ctx->sender_retry_queue[index];
 			if (retry->seq == seq && retry->peer == peer) {
 				delta = (now - retry->insert_time) / RIST_CLOCK;
+				/* this retry hasn't been handled yet, it makes no sense to insert a duplicate */
+				if (retry->active)
+					return;
 				if (delta < rtt)
 				{
 					rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 						"Nack request for seq %" PRIu32 " with delta %" PRIu64 "ms (age %"PRIu64"ms) is already queued (too soon to add another one), skipped, peer #%d '%s'\n",
-						buffer->seq, delta, age_ticks / RIST_CLOCK, peer->adv_peer_id, peer->receiver_name);
+						seq, delta, age_ticks / RIST_CLOCK, peer->adv_peer_id, peer->receiver_name);
 					peer->stats_sender_instant.bloat_skip++;
 					return;
 				}
@@ -1237,7 +1236,7 @@ void rist_retry_enqueue(struct rist_sender *ctx, uint32_t seq, struct rist_peer
 			if (ctx->common.debug) {
 				rist_log_priv(&ctx->common, RIST_LOG_DEBUG,
 					"Nack request for seq %" PRIu32 " with delta %" PRIu64 "ms (age %"PRIu64"ms) and rtt_min %" PRIu32 " for peer #%d '%s'\n",
-					buffer->seq, delta, age_ticks / RIST_CLOCK, peer->config.recovery_rtt_min, peer->adv_peer_id, peer->receiver_name);
+					seq, delta, age_ticks / RIST_CLOCK, peer->config.recovery_rtt_min, peer->adv_peer_id, peer->receiver_name);
 			}
 		}
 	}
@@ -1247,6 +1246,7 @@ void rist_retry_enqueue(struct rist_sender *ctx, uint32_t seq, struct rist_peer
 	retry->seq = seq;
 	retry->peer = peer;
 	retry->insert_time = now;
+	retry->active = true;
 	if (++ctx->sender_retry_queue_write_index >= ctx->sender_retry_queue_size) {
 		ctx->sender_retry_queue_write_index = 0;
 	}
