@@ -7,7 +7,8 @@
 #include <librist/librist.h>
 #include <librist/udpsocket.h>
 #include "librist/version.h"
-#ifdef USE_MBEDTLS
+#include "config.h"
+#if HAVE_MBEDTLS
 #include "librist/librist_srp.h"
 #include "srp_shared.h"
 #endif
@@ -40,6 +41,7 @@
 #define DATA_READ_MODE_POLL 1
 #define DATA_READ_MODE_API 2
 
+pthread_mutex_t signal_lock;
 static int signalReceived = 0;
 static struct rist_logging_settings logging_settings = LOGGING_SETTINGS_INITIALIZER;
 enum rist_profile profile = RIST_PROFILE_MAIN;
@@ -56,7 +58,7 @@ static struct option long_options[] = {
 { "stats",           required_argument, NULL, 'S' },
 { "verbose-level",   required_argument, NULL, 'v' },
 { "remote-logging",  required_argument, NULL, 'r' },
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 { "srpfile",         required_argument, NULL, 'F' },
 #endif
 { "help",            no_argument,       NULL, 'h' },
@@ -74,7 +76,7 @@ const char help_str[] = "Usage: %s [OPTIONS] \nWhere OPTIONS are:\n"
 "       -S | --statsinterval value (ms)           | Interval at which stats get printed, 0 to disable        |\n"
 "       -v | --verbose-level value                | To disable logging: -1, log levels match syslog levels   |\n"
 "       -r | --remote-logging IP:PORT             | Send logs and stats to this IP:PORT using udp messages   |\n"
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 "       -F | --srpfile filepath                   | When in listening mode, use this file to hold the list   |\n"
 "                                                 | of usernames and passwords to validate against. Use the  |\n"
 "                                                 | ristsrppasswd tool to create the line entries.           |\n"
@@ -95,7 +97,7 @@ static void usage(char *cmd)
 
 struct rist_callback_object {
 	int mpeg[MAX_OUTPUT_COUNT];
-	const struct rist_udp_config *udp_config[MAX_OUTPUT_COUNT];
+	struct rist_udp_config *udp_config[MAX_OUTPUT_COUNT];
 	uint16_t i_seqnum[MAX_OUTPUT_COUNT];
 };
 
@@ -133,7 +135,7 @@ static void connection_status_callback(void *arg, struct rist_peer *peer, enum r
 				peer, peer_connection_status, peer_connected_count);
 }
 
-static int cb_recv(void *arg, const struct rist_data_block *b)
+static int cb_recv(void *arg, struct rist_data_block *b)
 {
 	struct rist_callback_object *callback_object = (void *) arg;
 
@@ -142,7 +144,7 @@ static int cb_recv(void *arg, const struct rist_data_block *b)
 	for (i = 0; i < MAX_OUTPUT_COUNT; i++) {
 		if (!callback_object->udp_config[i])
 			continue;
-		const struct rist_udp_config *udp_config = callback_object->udp_config[i];
+		struct rist_udp_config *udp_config = callback_object->udp_config[i];
 		// The stream-id on the udp url gets translated into the virtual destination port of the GRE tunnel
 		uint16_t virt_dst_port = udp_config->stream_id;
 		// look for the correct mapping of destination port to output
@@ -182,13 +184,14 @@ static int cb_recv(void *arg, const struct rist_data_block *b)
 		rist_log(&logging_settings, RIST_LOG_ERROR, "Destination port mismatch, no output found for %d\n", b->virt_dst_port);
 		return -1;
 	}
-	rist_receiver_data_block_free((struct rist_data_block **const) &b);
+	rist_receiver_data_block_free2(&b);
 	return 0;
 }
 
 static void intHandler(int signal) {
-	rist_log(&logging_settings, RIST_LOG_INFO, "Signal %d received\n", signal);
+	pthread_mutex_lock(&signal_lock);
 	signalReceived = signal;
+	pthread_mutex_unlock(&signal_lock);
 }
 
 static int cb_auth_connect(void *arg, const char* connecting_ip, uint16_t connecting_port, const char* local_ip, uint16_t local_port, struct rist_peer *peer)
@@ -284,12 +287,17 @@ int main(int argc, char *argv[])
 	enum rist_log_level loglevel = RIST_LOG_INFO;
 	int statsinterval = 1000;
 	char *remote_log_address = NULL;
+	if (pthread_mutex_init(&signal_lock, NULL) != 0)
+	{
+		fprintf(stderr, "Could not initialize signal lock\n");
+		exit(1);
+	}
 #ifndef _WIN32
 	/* Receiver pipe handle */
 	int receiver_pipe[2];
 #endif
 
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 	FILE *srpfile = NULL;
 #endif
 
@@ -298,18 +306,16 @@ int main(int argc, char *argv[])
 		callback_object.mpeg[i] = 0;
 		callback_object.udp_config[i] = NULL;
 	}
-
 #ifdef _WIN32
 #define STDERR_FILENO 2
-    signal(SIGINT, intHandler);
-    signal(SIGTERM, intHandler);
-    signal(SIGABRT, intHandler);
+	signal(SIGINT, intHandler);
+	signal(SIGTERM, intHandler);
+	signal(SIGABRT, intHandler);
 #else
 	struct sigaction act = { {0} };
 	act.sa_handler = intHandler;
 	sigaction(SIGINT, &act, NULL);
 #endif
-
 	// Default log settings
     struct rist_logging_settings *log_ptr = &logging_settings;
     if (rist_logging_set(&log_ptr, loglevel, NULL, NULL, NULL,
@@ -352,7 +358,7 @@ int main(int argc, char *argv[])
 		case 'r':
 			remote_log_address = strdup(optarg);
 		break;
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 		case 'F':
 			srpfile = fopen(optarg, "r");
 			if (!srpfile) {
@@ -424,15 +430,15 @@ int main(int argc, char *argv[])
 			break;
 
 		// Rely on the library to parse the url
-		const struct rist_peer_config *peer_config = NULL;
-		if (rist_parse_address(inputtoken, (void *)&peer_config))
+		struct rist_peer_config *peer_config = NULL;
+		if (rist_parse_address2(inputtoken, &peer_config))
 		{
 			rist_log(&logging_settings, RIST_LOG_ERROR, "Could not parse peer options for receiver #%d\n", (int)(i + 1));
 			exit(1);
 		}
 
 		/* Process overrides */
-		struct rist_peer_config *overrides_peer_config = (void *)peer_config;
+		struct rist_peer_config *overrides_peer_config = peer_config;
 		if (shared_secret && peer_config->secret[0] == 0) {
 			strncpy(overrides_peer_config->secret, shared_secret, RIST_MAX_STRING_SHORT -1);
 			if (encryption_type)
@@ -458,7 +464,7 @@ int main(int argc, char *argv[])
 			rist_log(&logging_settings, RIST_LOG_ERROR, "Could not add peer connector to receiver #%i\n", (int)(i + 1));
 			exit(1);
 		}
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 		int srp_error = 0;
 		if (profile != RIST_PROFILE_SIMPLE) {
 			if (strlen(peer_config->srp_username) > 0 && strlen(peer_config->srp_password) > 0)
@@ -478,7 +484,7 @@ int main(int argc, char *argv[])
 			rist_log(&logging_settings, RIST_LOG_WARN, "SRP Authentication is not available for Rist Simple Profile\n");
 #endif
 
-		rist_peer_config_free(&peer_config);
+		rist_peer_config_free2(&peer_config);
 		inputtoken = strtok_r(NULL, ",", &saveptr1);
 	}
 
@@ -495,8 +501,8 @@ int main(int argc, char *argv[])
 		// We are using the rist_parse_address function to create a config object that does not really
 		// belong to the udp output. We do this only to avoid writing another parser for the two url
 		// parameters available to the udp input/output url
-		const struct rist_udp_config *udp_config = NULL;
-		if (rist_parse_udp_address(outputtoken, &udp_config)) {
+		struct rist_udp_config *udp_config = NULL;
+		if (rist_parse_udp_address2(outputtoken, &udp_config)) {
 			rist_log(&logging_settings, RIST_LOG_ERROR, "Could not parse outputurl %s\n", outputtoken);
 			goto next;
 		}
@@ -513,7 +519,7 @@ int main(int argc, char *argv[])
 
 		// Open the output socket
 		callback_object.mpeg[i] = udpsocket_open_connect(hostname, outputport, udp_config->miface);
-		if (callback_object.mpeg[i] <= 0) {
+		if (callback_object.mpeg[i] < 0) {
 			rist_log(&logging_settings, RIST_LOG_ERROR, "Could not connect to: Host %s, Port %d\n", (char *) hostname, outputport);
 			goto next;
 		} else {
@@ -530,11 +536,8 @@ next:
 		exit(1);
 	}
 
-	// callback is best unless you are using the timestamps passed with the buffer
-	data_read_mode = DATA_READ_MODE_CALLBACK;
-
 	if (data_read_mode == DATA_READ_MODE_CALLBACK) {
-		if (rist_receiver_data_callback_set(ctx, cb_recv, &callback_object))
+		if (rist_receiver_data_callback_set2(ctx, cb_recv, &callback_object))
 		{
 			rist_log(&logging_settings, RIST_LOG_ERROR, "Could not set data_callback pointer\n");
 			exit(1);
@@ -588,10 +591,10 @@ next:
 		SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
 #endif
 		// Master loop
-		while (!signalReceived)
+		while (true)
 		{
-			const struct rist_data_block *b = NULL;
-			int queue_size = rist_receiver_data_read(ctx, &b, 5);
+			struct rist_data_block *b = NULL;
+			int queue_size = rist_receiver_data_read2(ctx, &b, 5);
 			if (queue_size > 0) {
 				if (queue_size % 10 == 0 || queue_size > 50) {
 					// We need a better way to report on this
@@ -600,6 +603,13 @@ next:
 				}
 				if (b && b->payload) cb_recv(&callback_object, b);
 			}
+			pthread_mutex_lock(&signal_lock);
+			if (signalReceived)
+			{
+				rist_log(&logging_settings, RIST_LOG_INFO, "Signal %d received\n", signal);
+				break;
+			}
+			pthread_mutex_unlock(&signal_lock);
 		}
 	}
 #ifndef _WIN32
@@ -609,7 +619,7 @@ next:
 		struct timeval timeout;
 		timeout.tv_sec = 0;
 		timeout.tv_usec = 5000;
-		while (!signalReceived) {
+		while (true) {
 			FD_ZERO(&readfds);
 			FD_SET(receiver_pipe[ReadEnd], &readfds);
 			int ret = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout);
@@ -630,10 +640,10 @@ next:
 				}
 			}
 			/* Consume data from library */
-			const struct rist_data_block *b = NULL;
+			struct rist_data_block *b = NULL;
 			int queue_size = 0;
 			for (;;) {
-				queue_size = rist_receiver_data_read(ctx, &b, 0);
+				queue_size = rist_receiver_data_read2(ctx, &b, 0);
 				if (queue_size > 0) {
 					if (queue_size % 10 == 0 || queue_size > 50) {
 						// We need a better way to report on this
@@ -645,16 +655,23 @@ next:
 				else
 					break;
 			}
+			pthread_mutex_lock(&signal_lock);
+			if (signalReceived)
+			{
+				rist_log(&logging_settings, RIST_LOG_INFO, "Signal %d received\n", signal);
+				break;
+			}
+			pthread_mutex_unlock(&signal_lock);
 		}
 	}
 #endif
-
+	fprintf(stderr, "DESTROY\n");
 	rist_destroy(ctx);
 
 	for (size_t i = 0; i < MAX_OUTPUT_COUNT; i++) {
 		// Free udp_config object
 		if ((void *)callback_object.udp_config[i])
-			rist_udp_config_free(&callback_object.udp_config[i]);
+			rist_udp_config_free2(&callback_object.udp_config[i]);
 	}
 
 	rist_logging_unset_global();
