@@ -7,7 +7,8 @@
 #include <librist/librist.h>
 #include <librist/udpsocket.h>
 #include "librist/version.h"
-#ifdef USE_MBEDTLS
+#include "config.h"
+#if HAVE_MBEDTLS
 #include "librist/librist_srp.h"
 #include "srp_shared.h"
 #endif
@@ -47,14 +48,14 @@ struct rist_callback_object {
 	struct evsocket_ctx *evctx;
 	struct rist_ctx *receiver_ctx;
 	struct rist_ctx *sender_ctx;
-	const struct rist_udp_config *udp_config;
+	struct rist_udp_config *udp_config;
 	uint8_t recv[RIST_MAX_PACKET_SIZE];
 };
 
 struct receive_thread_object {
 	int sd;
 	struct rist_ctx *ctx[MAX_OUTPUT_COUNT];
-	const struct rist_udp_config *udp_config;
+	struct rist_udp_config *udp_config;
 	uint8_t recv[RIST_MAX_PACKET_SIZE];
 };
 
@@ -82,7 +83,7 @@ static struct option long_options[] = {
 { "stats",           required_argument, NULL, 'S' },
 { "verbose-level",   required_argument, NULL, 'v' },
 { "remote-logging",  required_argument, NULL, 'r' },
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 { "srpfile",         required_argument, NULL, 'F' },
 #endif
 { "fast-start",      required_argument, NULL, 'f' },
@@ -102,7 +103,7 @@ const char help_str[] = "Usage: %s [OPTIONS] \nWhere OPTIONS are:\n"
 "       -S | --statsinterval value (ms)           | Interval at which stats get printed, 0 to disable        |\n"
 "       -v | --verbose-level value                | To disable logging: -1, log levels match syslog levels   |\n"
 "       -r | --remote-logging IP:PORT             | Send logs and stats to this IP:PORT using udp messages   |\n"
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 "       -F | --srpfile filepath                   | When in listening mode, use this file to hold the list   |\n"
 "                                                 | of usernames and passwords to validate against. Use the  |\n"
 "                                                 | ristsrppasswd tool to create the line entries.           |\n"
@@ -130,7 +131,7 @@ static uint64_t risttools_convertRTPtoNTP(uint32_t i_rtp)
 }
 */
 
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 	FILE *srpfile = NULL;
 #endif
 
@@ -305,8 +306,8 @@ static struct rist_peer* setup_rist_peer(struct rist_sender_args *setup)
 	}
 
 	// Rely on the library to parse the url
-	const struct rist_peer_config *peer_config_link = NULL;
-	if (rist_parse_address(setup->token, (void *)&peer_config_link))
+	struct rist_peer_config *peer_config_link = NULL;
+	if (rist_parse_address2(setup->token, (void *)&peer_config_link))
 	{
 		rist_log(&logging_settings, RIST_LOG_ERROR, "Could not parse peer options for sender: %s\n", setup->token);
 		return NULL;
@@ -348,7 +349,7 @@ static struct rist_peer* setup_rist_peer(struct rist_sender_args *setup)
 		return NULL;
 	}
 
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 	int srp_error = 0;
 	if (setup->profile != RIST_PROFILE_SIMPLE) {
 		if (strlen(peer_config_link->srp_username) > 0 && strlen(peer_config_link->srp_password) > 0)
@@ -368,7 +369,7 @@ static struct rist_peer* setup_rist_peer(struct rist_sender_args *setup)
 		rist_log(&logging_settings, RIST_LOG_WARN, "SRP Authentication is not available for Rist Simple Profile\n");
 #endif
 
-	rist_peer_config_free(&peer_config_link);
+	rist_peer_config_free2(&peer_config_link);
 
 	return peer;
 }
@@ -381,8 +382,8 @@ static PTHREAD_START_FUNC(input_loop, arg)
 		if (callback_object->receiver_ctx)
 		{
 			// RIST receiver
-			const struct rist_data_block *b;
-			int queue_size = rist_receiver_data_read(callback_object->receiver_ctx, &b, 5);
+			struct rist_data_block *b = NULL;
+			int queue_size = rist_receiver_data_read2(callback_object->receiver_ctx, &b, 5);
 			if (queue_size > 0) {
 				if (queue_size % 10 == 0 || queue_size > 50)
 					rist_log(&logging_settings, RIST_LOG_WARN, "Falling behind on rist_receiver_data_read: %d\n", queue_size);
@@ -392,7 +393,7 @@ static PTHREAD_START_FUNC(input_loop, arg)
 						// TODO: report error?
 						(void) w;
 					}
-					rist_receiver_data_block_free((struct rist_data_block **const)&b);
+					rist_receiver_data_block_free2(&b);
 				}
 			}
 		}
@@ -424,6 +425,8 @@ int main(int argc, char *argv[])
 	int faststart = 0;
 	struct rist_sender_args peer_args;
 	char *remote_log_address = NULL;
+	bool thread_started[MAX_INPUT_COUNT +1] = {false};
+	pthread_t thread_main_loop[MAX_INPUT_COUNT+1] = { 0 };
 
 	for (size_t i = 0; i < MAX_INPUT_COUNT; i++)
 		event[i] = NULL;
@@ -481,7 +484,7 @@ int main(int argc, char *argv[])
 		case 'r':
 			remote_log_address = strdup(optarg);
 		break;
-#ifdef USE_MBEDTLS
+#if HAVE_MBEDTLS
 		case 'F':
 			srpfile = fopen(optarg, "r");
 			if (!srpfile) {
@@ -545,23 +548,24 @@ int main(int argc, char *argv[])
 	bool atleast_one_socket_opened = false;
 	char *saveptrinput;
 	char *inputtoken = strtok_r(inputurl, ",", &saveptrinput);
+	struct rist_udp_config *udp_config = NULL;
 	for (size_t i = 0; i < MAX_INPUT_COUNT; i++) {
 		if (!inputtoken)
 			break;
 
 		// First parse extra url and parameters
-		const struct rist_udp_config *udp_config = NULL;
-		if (rist_parse_udp_address(inputtoken, &udp_config)) {
+		if (rist_parse_udp_address2(inputtoken, &udp_config)) {
 			rist_log(&logging_settings, RIST_LOG_ERROR, "Could not parse inputurl %s\n", inputtoken);
 			goto next;
 		}
 
 		// Check for duplicate stream-ids and reject the entire config if we have any dups
+		bool found_empty = false;
 		for (size_t j = 0; j < MAX_INPUT_COUNT; j++) {
-			if (stream_id_check[j] == -1) {
+			if (stream_id_check[j] == -1 && !found_empty) {
 				stream_id_check[j] = (int32_t)udp_config->stream_id;
 				rist_log(&logging_settings, RIST_LOG_INFO, "Assigning stream-id %d to this input\n", udp_config->stream_id);
-				break;
+				found_empty = true;
 			} else if ((uint16_t)stream_id_check[j] == udp_config->stream_id) {
 				rist_log(&logging_settings, RIST_LOG_ERROR, "Every input must have a unique stream-id (%d) when you multiplex\n", udp_config->stream_id);
 				goto shutdown;
@@ -612,6 +616,8 @@ int main(int argc, char *argv[])
 			struct rist_peer *peer = setup_rist_peer(&peer_args);
 			if (peer == NULL)
 				atleast_one_socket_opened = true;
+			rist_udp_config_free2(&udp_config);
+			udp_config = NULL;
 		}
 		else {
 			if(!evctx)
@@ -627,7 +633,7 @@ int main(int argc, char *argv[])
 			rist_log(&logging_settings, RIST_LOG_INFO, "URL parsed successfully: Host %s, Port %d\n", (char *) hostname, inputport);
 
 			callback_object[i].sd = udpsocket_open_bind(hostname, inputport, udp_config->miface);
-			if (callback_object[i].sd <= 0) {
+			if (callback_object[i].sd < 0) {
 				rist_log(&logging_settings, RIST_LOG_ERROR, "Could not bind to: Host %s, Port %d, miface %s.\n",
 					(char *) hostname, inputport, udp_config->miface);
 				goto next;
@@ -637,7 +643,7 @@ int main(int argc, char *argv[])
 				atleast_one_socket_opened = true;
 			}
 			callback_object[i].udp_config = udp_config;
-
+			udp_config = NULL;
 			callback_object[i].evctx = evctx;
 			event[i] = evsocket_addevent(callback_object[i].evctx, callback_object[i].sd, EVSOCKET_EV_READ, input_udp_recv, input_udp_sockerr,
 				(void *)&callback_object[i]);
@@ -651,13 +657,12 @@ next:
 		goto shutdown;
 	}
 
-	pthread_t thread_main_loop[MAX_INPUT_COUNT+1] = { 0 };
-
 	if (evctx && pthread_create(&thread_main_loop[0], NULL, input_loop, (void *)callback_object) != 0)
 	{
 		fprintf(stderr, "Could not start udp receiver thread\n");
 		goto shutdown;
 	}
+	thread_started[0] = true;
 
 	for (size_t i = 0; i < MAX_INPUT_COUNT; i++) {
 		if (callback_object[i].sender_ctx && rist_start(callback_object[i].sender_ctx) == -1) {
@@ -673,6 +678,7 @@ next:
 			fprintf(stderr, "Could not start send rist receiver thread\n");
 			goto shutdown;
 		}
+		thread_started[i+1] = true;
 	}
 
 #ifdef _WIN32
@@ -682,13 +688,16 @@ next:
 #endif
 
 shutdown:
+	if (udp_config) {
+		rist_udp_config_free2(&udp_config);
+	}
 	for (size_t i = 0; i < MAX_INPUT_COUNT; i++) {
 		// Remove socket events
 		if (event[i])
 			evsocket_delevent(callback_object[i].evctx, event[i]);
 		// Free udp_config object
 		if ((void *)callback_object[i].udp_config)
-			rist_udp_config_free(&callback_object[i].udp_config);
+			rist_udp_config_free2(&callback_object[i].udp_config);
 		// Cleanup rist listeners
 		if (callback_object[i].receiver_ctx)
 			rist_destroy(callback_object[i].receiver_ctx);
@@ -698,7 +707,7 @@ shutdown:
 	}
 
 	for (size_t i = 0; i <= MAX_INPUT_COUNT; i++) {
-		if (thread_main_loop[i])
+		if (thread_started[i])
 			pthread_join(thread_main_loop[i], NULL);
 	}
 
